From gpfsl.orc11 Require Export base.
Require Import iris.bi.lib.fractional.
From lrust.lifetime Require Export lifetime_sig.
From lrust.lifetime.model Require definitions primitive accessors faking borrow
     borrow_sep reborrow creation in_at_borrow.
From iris.proofmode Require Import tactics.

From iris.prelude Require Import options.

Module Export lifetime : lifetime_sig.
  Definition lft := gmultiset positive.
  Include definitions.
  Include primitive.
  Include borrow.
  Include faking.
  Include borrow_sep.
  Include reborrow.
  Include accessors.
  Include creation.
  Include in_at_borrow.
End lifetime.

Notation lft_intersect_list l := (foldr (⊓) static l).

Global Instance lft_inhabited : Inhabited lft := populate static.

Canonical lftO := leibnizO lft.

Definition lft_incl_syntactic (κ κ' : lft) : Prop := ∃ κ'', κ'' ⊓ κ' = κ.
Infix "⊑ˢʸⁿ" := lft_incl_syntactic (at level 40) : stdpp_scope.

Section derived.
Context `{!invGS Σ, !lftG Σ Lat userE}.
Implicit Types κ : lft.

(* TODO: upstream *)
Global Instance lft_mul_fractional q κ : Fractional (λ q', (q * q').[κ])%I.
Proof. intros ??. by rewrite Qp.mul_add_distr_l lft_tok_fractional. Qed.

Lemma lft_create E `{LatBottom Lat} :
  ↑lftN ⊆ E →
   ⎡lft_ctx⎤ ={E}=∗ ∃ κ, 1.[κ] ∗ □ (1.[κ] ={↑lftN ∪ userE}[userE]▷=∗ [†κ]).
Proof.
  iIntros (?) "#LFT".
  iMod (lft_create_strong (λ _, True) with "LFT") as (κ _) "H"=>//;
    [by apply pred_infinite_True|].
  by auto.
Qed.

Lemma lft_tok_dead_subj q κ :
  q.[κ] -∗ <subj> [† κ] -∗ False.
Proof.
  iStartProof (iProp _). iIntros "% ?" (V) "-> H†".
  iDestruct "H†" as (?) "H†".
  iApply (lft_tok_dead $! (_⊔_) with "[$] [$]").
Qed.

Lemma bor_shorten κ κ' P :
  κ ⊑ κ' -∗ &{κ'}P -∗ &{κ}P.
Proof.
  rewrite !bor_unfold_idx. iIntros "#?".
  iDestruct 1 as (i) "[??]". iExists i. iFrame. by iApply idx_bor_shorten.
Qed.

Lemma bor_acc_atomic_cons E κ P :
  ↑lftN ⊆ E →
  ⎡lft_ctx⎤ -∗ &{κ}P ={E,E∖↑lftN}=∗
    (∃ P', ▷ (⎡P'⎤ ∧ P) ∗ ∀ Q,
            ▷ (▷ Q ={userE}=∗ ▷ P) -∗ ▷ (⎡P'⎤ ∧ Q) ={E∖↑lftN,E}=∗ &{κ}Q) ∨
    (∃ κ', κ ⊑ κ' ∗ <subj> [†κ'] ∗ |={E∖↑lftN,E}=> True).
Proof.
  iIntros (?) "#LFT HP".
  iMod (bor_acc_atomic_strong with "LFT HP") as "[H|H]"; [done| |auto].
  iLeft. iDestruct "H" as (P' κ') "(#Hκκ' & H & Hclose)". iExists P'.
  iIntros "{$H} !> % HPQ HQ". iMod ("Hclose" with "[HPQ] HQ") as "Hb".
  - iNext. iIntros "? _". by iApply "HPQ".
  - iApply (bor_shorten with "Hκκ' Hb").
Qed.

Lemma bor_acc_atomic E κ P :
  ↑lftN ⊆ E →
  ⎡lft_ctx⎤ -∗ &{κ}P ={E,E∖↑lftN}=∗
       (∃ P', ▷ (⎡P'⎤ ∧ P) ∗ (▷ ⎡P'⎤ ={E∖↑lftN,E}=∗ &{κ}P)) ∨
       (∃ κ', κ ⊑ κ' ∗ <subj> [†κ'] ∗ |={E∖↑lftN,E}=> True).
Proof.
  iIntros (?) "#LFT HP".
  iMod (bor_acc_atomic_cons with "LFT HP") as "[H|H]"; [done| |auto].
  iDestruct "H" as (P') "[HP Hclose]".
  iDestruct (monPred_in_intro with "HP") as (V) "[#HV HP]".
  iLeft. iExists (P' ∧ P V)%I. iSplitL "HP".
  - iModIntro. iSplit; [by iNext|]. iApply monPred_in_elim. { done. }
    iDestruct "HP" as "[_ HP]". by iNext.
  - iIntros "!> HP". iMod ("Hclose" with "[] [HP]") as "$"; [by iIntros "!> $"| |done].
    iSplit.
    + iDestruct "HP" as "[$ _]".
    + iApply monPred_in_elim. { done. }
      iNext. iDestruct "HP" as "[_ $]".
Qed.

Lemma bor_acc_cons E q κ P :
  ↑lftN ⊆ E →
  ⎡lft_ctx⎤ -∗ &{κ}P -∗ q.[κ] ={E}=∗ ▷ P ∗
    ∀ Q, ▷ (▷ Q ={userE}=∗ ▷ P) -∗ ▷ Q ={E}=∗ &{κ}Q ∗ q.[κ].
Proof.
  iIntros (?) "#LFT HP Htok".
  iMod (bor_acc_strong with "LFT HP Htok") as (κ') "(#Hκκ' & $ & Hclose)"; first done.
  iIntros "!> * HPQ HQ". iMod ("Hclose" with "[HPQ] HQ") as "[Hb $]".
  - iNext. iIntros "? _". by iApply "HPQ".
  - iApply (bor_shorten with "Hκκ' Hb").
Qed.

Lemma bor_acc E q κ P :
  ↑lftN ⊆ E →
  ⎡lft_ctx⎤ -∗ &{κ}P -∗ q.[κ] ={E}=∗ ▷ P ∗ (▷ P ={E}=∗ &{κ}P ∗ q.[κ]).
Proof.
  iIntros (?) "#LFT HP Htok".
  iMod (bor_acc_cons with "LFT HP Htok") as "($ & Hclose)"; [done|].
  iIntros "!> HP". iMod ("Hclose" with "[] HP") as "[$ $]"; auto.
Qed.

Lemma idx_bor_acc_sync E q κ i P :
  ↑lftN ⊆ E →
  ⎡lft_ctx⎤ -∗ &{κ,i}P -∗ idx_bor_own i -∗ q.[κ] ={E}=∗
      ▷ P ∗ (▷ P ={E}=∗ idx_bor_own i ∗ q.[κ]).
Proof.
  iIntros (?) "#LFT Hb Hown Htok". iCombine "Hb Hown Htok" as "HH".
  iDestruct (monPred_in_intro with "HH") as (V) "(#HV & #Hb & Hown & Htok)".
  iMod (idx_bor_acc with "LFT Hb Hown Htok") as (V' HV') "(HP & Hclose)". { done. }
  revert HV'. rewrite lat_join_idem=>HV'. iSplitL "HP".
  - iModIntro. iApply monPred_in_elim; [|by iNext]. by rewrite HV'.
  - iIntros "!> HP". iCombine "HV HP" as"HH".
    iDestruct (monPred_in_intro with "HH") as (V'') "(#HV'' & % & HP)".
    iMod ("Hclose" $! V'' with "[//] [$HP]") as "[Hown Htok]".
    iApply monPred_in_elim; [done|]. by iFrame.
Qed.

Lemma bor_exists `{LatBottom Lat, !Inhabited A} (Φ : A → monPred _ _) E κ :
  ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ}(∃ x, Φ x) ={E}=∗ ∃ x, &{κ}(Φ x).
Proof.
  iIntros (?) "#LFT Hb".
  iMod (bor_acc_atomic_cons with "LFT Hb") as "[H|H]"; [done| |].
  - iDestruct "H" as (P') "[HΦ Hclose]". rewrite bi.and_exist_l.
    iDestruct "HΦ" as (x) "HΦ". iExists x.
    iApply ("Hclose" with "[] HΦ"). iIntros "!> ?". eauto.
  - iExists inhabitant. iDestruct "H" as (κ') "(#? & #? & >_)".
    by iApply bor_shorten; [|iApply bor_fake].
Qed.

Lemma bor_or `{LatBottom Lat} E κ P Q :
  ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ}(P ∨ Q) ={E}=∗ &{κ}P ∨ &{κ}Q.
Proof.
  iIntros (?) "#LFT H". rewrite bi.or_alt.
  (* The (A:=...) is needed due to Coq bug #5458 *)
  iMod (bor_exists (A:=bool) with "LFT H") as ([]) "H"; auto.
Qed.

Lemma bor_iff κ P P' : ▷ □ (P ↔ P') -∗ &{κ}P -∗ &{κ}P'.
Proof.
  rewrite !bor_unfold_idx. iIntros "#HP Hbor".
  iDestruct "Hbor" as (i) "[Hbor Htok]". iExists i. iFrame "Htok".
  iApply idx_bor_iff; done.
Qed.

Lemma bor_iff_proper κ P P': ▷ □ (P ↔ P') -∗ □ (&{κ}P ↔ &{κ}P').
Proof.
  iIntros "#HP !>". iSplit; iIntros "?"; iApply bor_iff; try done.
  iNext. iModIntro. iSplit; iIntros "?"; iApply "HP"; done.
Qed.

Lemma bor_later `{LatBottom Lat} E κ P :
  ↑lftN ⊆ E →
  ⎡lft_ctx⎤ -∗ &{κ}(▷ P) ={E}[E∖↑lftN]▷=∗ &{κ}P.
Proof.
  iIntros (?) "#LFT Hb".
  iMod (bor_acc_atomic_cons with "LFT Hb") as "[H|H]"; first done.
  - iDestruct "H" as (P') "[HP  Hclose]". iModIntro. iNext.
    iApply ("Hclose" with "[] [HP]"); [by iIntros "!> $"|]. by iNext.
  - iIntros "!> !>". iDestruct "H" as (κ') "(#? & #? & >_)".
    by iApply bor_shorten; [|iApply bor_fake].
Qed.

Lemma bor_later_tok E q κ P :
  ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ}(▷ P) -∗ q.[κ] ={E}▷=∗ &{κ}P ∗ q.[κ].
Proof.
  iIntros (?) "#LFT Hb Htok".
  iMod (bor_acc_cons with "LFT Hb Htok") as "[HP Hclose]"; first done.
  iModIntro. iNext. iApply ("Hclose" with "[] HP"). by iIntros "!> $".
Qed.

Lemma bor_persistent_notok P `{!Persistent P} E κ :
  ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ}P ={E}=∗ ▷ P ∨ ∃ κ', κ ⊑ κ' ∗ <subj> [†κ'].
Proof.
  iIntros (?) "#LFT Hb".
  iMod (bor_acc_atomic with "LFT Hb") as "[H| H]"; [done| |].
  - iLeft. iDestruct "H" as (P') "[HP Hob]". iDestruct "HP" as "[HP $]".
    by iMod ("Hob" with "HP") as "_".
  - iDestruct "H" as (?) "(? & ? & >_)". auto.
Qed.

Lemma bor_persistent P `{!Persistent P} E κ q :
  ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ}P -∗ q.[κ] ={E}=∗ ▷ P ∗ q.[κ].
Proof.
  iIntros (?) "#LFT Hb Htok".
  iMod (bor_acc with "LFT Hb Htok") as "[#HP Hob]"; [done|].
  by iMod ("Hob" with "HP") as "[_ $]".
Qed.

Lemma later_bor_static `{LatBottom Lat} E P :
  ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ ▷ P ={E}=∗ &{static}P.
Proof. iIntros (?) "#LFT HP". by iMod (bor_create with "LFT HP") as "[$ _]". Qed.

Lemma bor_static_later E P :
  ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{static}P ={E}=∗ ▷ P.
Proof.
  iIntros (?) "LFT HP". iMod (bor_acc _ 1%Qp with "LFT HP []") as "[$ _]"=>//.
  iApply lft_tok_static.
Qed.

Lemma rebor `{LatBottom Lat} E κ κ' P :
  ↑lftN ⊆ E →
  ⎡lft_ctx⎤ -∗ κ' ⊑ κ -∗ &{κ}P ={E}=∗ &{κ'}P ∗ ([†κ'] ={E}=∗ &{κ}P).
Proof.
  iIntros (?) "#LFT #Hκ'κ Hbor". setoid_rewrite ->(bor_unfold_idx κ).
  iDestruct "Hbor" as (i) "[#Hbor Hidx]".
  iMod (bor_create _ κ' with "LFT Hidx") as "[Hidx Hinh]"; first done.
  iMod (idx_bor_unnest with "LFT Hbor Hidx") as "Hbor'"; first done.
  iDestruct (bor_shorten with "[] Hbor'") as "$".
  { iApply lft_incl_glb; [done|]. iApply lft_incl_refl. }
  iIntros "!> H†". iMod ("Hinh" with "H†") as ">Hidx". auto.
Qed.

Lemma bor_unnest `{LatBottom Lat} E κ κ' P :
  ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ'}(&{κ}P) ={E}▷=∗ &{κ ⊓ κ'}P.
Proof.
  iIntros (?) "#LFT Hbor". rewrite ->(bor_unfold_idx _ P).
  iMod (bor_exists (A:=bor_idx) with "LFT Hbor") as (i) "Hbor"; [done|].
  iMod (bor_sep with "LFT Hbor") as "[Hidx Hbor]"; [done|].
  iMod (bor_persistent_notok with "LFT Hidx") as "#[Hidx|H†]"; [done| |].
  - iIntros "!>". iNext. by iApply (idx_bor_unnest with "LFT Hidx Hbor").
  - iDestruct "H†" as (κ'') "[#? #?]".
    iApply bor_shorten; [|by iApply (bor_fake with "LFT")].
    iApply lft_incl_trans; [|done]. iApply lft_intersect_incl_r.
Qed.

Lemma lft_incl_static κ : ⊢ κ ⊑ static.
Proof.
  iApply lft_incl_intro. iIntros "!>" (q) "$". iExists 1%Qp. iSplitR; [|auto].
    by iApply lft_tok_static.
Qed.

Lemma lft_intersect_list_elem_of_incl κs κ :
  κ ∈ κs → ⊢ lft_intersect_list κs ⊑ κ.
Proof.
  induction 1 as [|???? IH].
  - iApply lft_intersect_incl_l.
  - iApply lft_incl_trans; last iApply IH.
    (* FIXME: Why does "done" not do this? Looks like "assumption" to me. *)
    iApply lft_intersect_incl_r.
Qed.

Lemma lft_eternalize E q κ : q.[κ] ={E}=∗ static ⊑ κ.
Proof.
  iIntros "Hκ". iDestruct (monPred_in_intro with "Hκ") as (V) "[#HV Hκ]".
  iMod (inv_alloc lftN _ (∃ q, q.[κ] V)%I with "[Hκ]") as "#Hnv".
  { iExists _. by iNext. }
  iApply lft_incl_intro. iIntros "!> !>" (q') "_".
  iInv lftN as ">Hκ" "Hclose". iDestruct "Hκ" as (q'') "[Hκ1 Hκ2]".
  iMod ("Hclose" with "[Hκ2]") as "_". { iExists _. by iNext. }
  iModIntro. iExists _. iSplitL; [by iApply monPred_in_elim; auto|].
  iIntros "_". iApply lft_tok_static.
Qed.

Lemma in_at_bor_fake `{LatBottom Lat} E κ P :
  ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ <subj> [†κ] ={E}=∗ &in_at{κ}P.
Proof.
  iIntros (?) "#LFT#H†". iApply (in_at_bor_share with "[>]"); try done.
  by iApply (bor_fake with "LFT H†").
Qed.

(** Syntactic lifetime inclusion *)
Lemma lft_incl_syn_sem κ κ' :
  κ ⊑ˢʸⁿ κ' → ⊢ κ ⊑ κ'.
Proof. intros [z Hxy]. rewrite -Hxy. by apply lft_intersect_incl_r. Qed.

Lemma lft_intersect_incl_syn_l κ κ' : κ ⊓ κ' ⊑ˢʸⁿ κ.
Proof. by exists κ'; rewrite (comm _ κ κ'). Qed.

Lemma lft_intersect_incl_syn_r κ κ' : κ ⊓ κ' ⊑ˢʸⁿ κ'.
Proof. by exists κ. Qed.

Lemma lft_incl_syn_refl κ : κ ⊑ˢʸⁿ κ.
Proof. exists static; by rewrite left_id. Qed.

Lemma lft_incl_syn_trans κ κ' κ'' :
  κ ⊑ˢʸⁿ κ' → κ' ⊑ˢʸⁿ κ'' → κ ⊑ˢʸⁿ κ''.
Proof.
  intros [κ0 Hκ0] [κ'0 Hκ'0].
  rewrite -Hκ0 -Hκ'0 assoc.
  by apply lft_intersect_incl_syn_r.
Qed.

Lemma lft_intersect_mono_syn κ1 κ1' κ2 κ2' :
  κ1 ⊑ˢʸⁿ κ1' → κ2 ⊑ˢʸⁿ κ2' → (κ1 ⊓ κ2) ⊑ˢʸⁿ (κ1' ⊓ κ2').
Proof.
  intros [κ1'' Hκ1] [κ2'' Hκ2].
  exists (κ1'' ⊓ κ2'').
  rewrite -!assoc (comm _ κ2'' _).
  rewrite -assoc assoc (comm _ κ2' _).
  by f_equal.
Qed.

Lemma lft_incl_syn_static κ : κ ⊑ˢʸⁿ static.
Proof.
  exists κ; by apply lft_intersect_right_id.
Qed.

Lemma lft_intersect_list_elem_of_incl_syn κs κ :
  κ ∈ κs → lft_intersect_list κs ⊑ˢʸⁿ κ.
Proof.
  induction 1 as [κ|κ κ' κs Hin IH].
  - apply lft_intersect_incl_syn_l.
  - eapply lft_incl_syn_trans; last done.
    apply lft_intersect_incl_syn_r.
Qed.

Global Instance lft_incl_syn_anti_symm : AntiSymm eq (λ x y, x ⊑ˢʸⁿ y).
Proof.
  intros κ1 κ2 [κ12 Hκ12] [κ21 Hκ21].
  assert (κ21 = static) as Hstatic.
  { apply (lft_intersect_static_cancel_l _ κ12).
    eapply (inj (κ1 ⊓.)).
    rewrite assoc right_id.
    eapply (inj (.⊓ κ2)).
    rewrite (comm _ κ1 κ21) -assoc Hκ21 Hκ12 (comm _ κ2). done. }
  by rewrite Hstatic left_id in Hκ21.
Qed.
End derived.
